"""
This class is necessary to be able to please the entrypoints of both localstack and AWS.

If we set the entrypoint directly to steamship.invocable.safe_handler (imported in the init
from lambda_handler), AWS is happy, but localstack is not because it tries to read steamship.invocable as a py file,
not a module.

If we set the entrypoint to steamship.invocable.lambda_handler.safe_handler, Localstack is happy, but AWS
is not, because it tries to read lambda_handler first, which imports things from steamship.invocable, which imports
things from lambda_handler.

By adding this file which basically no-ops safe_handler into steamship.invocable.entrypoint.safe_handler, both are
happy.

"""

from steamship.invocable import safe_handler

_ = safe_handler  # No op line so that my "unused" import does not get removed.



from __future__ import annotations

import logging
from enum import Enum
from typing import Any, List, Optional, Type

from pydantic import BaseModel, Field

from steamship.base.client import Client
from steamship.base.model import CamelModel
from steamship.base.request import GetRequest, IdentifierRequest
from steamship.base.request import Request
from steamship.base.request import Request as SteamshipRequest
from steamship.base.response import Response
from steamship.base.response import Response as SteamshipResponse


class ListWorkspacesRequest(Request):
    pass


class ListWorkspacesResponse(Response):
    workspaces: List[Workspace]


class Workspace(CamelModel):
    client: Client = Field(None, exclude=True)
    id: str = None
    handle: str = None

    @classmethod
    def parse_obj(cls: Type[BaseModel], obj: Any) -> BaseModel:
        # TODO (enias): This needs to be solved at the engine side\
        obj = obj["workspace"] if "workspace" in obj else obj
        return super().parse_obj(obj)

    class CreateRequest(SteamshipRequest):
        id: Optional[str] = None
        handle: Optional[str] = None
        fetch_if_exists: Optional[bool] = None
        external_id: Optional[str] = None
        external_type: Optional[str] = None
        metadata: Optional[str] = None

    def delete(self) -> Workspace:
        return self.client.post("workspace/delete", IdentifierRequest(id=self.id), expect=Workspace)

    @staticmethod
    def get(
        client: Client, id_: str = None, handle: str = None, fetch_if_exists: bool = None
    ) -> Workspace:
        req = GetRequest(id=id_, handle=handle, fetch_if_exists=fetch_if_exists)
        return client.post("workspace/get", req, expect=Workspace)

    @staticmethod
    def create(
        client: Client,
        handle: Optional[str] = None,
        external_id: Optional[str] = None,
        external_type: Optional[str] = None,
        metadata: Any = None,
        fetch_if_exists: bool = True,
    ) -> Workspace:
        req = Workspace.CreateRequest(
            handle=handle,
            fetch_if_exists=fetch_if_exists,
            external_id=external_id,
            external_type=external_type,
            metadata=metadata,
        )
        return client.post("workspace/create", req, expect=Workspace)

    def create_signed_url(self, request: SignedUrl.Request) -> SignedUrl.Response:
        logging.info(f"Requesting signed URL: {request}")
        ret = self.client.post(
            "workspace/createSignedUrl", payload=request, expect=SignedUrl.Response
        )
        logging.info(f"Got signed URL: {ret}")
        return ret

    @staticmethod
    def list(client: Client, t: str = None) -> ListWorkspacesResponse:
        return client.post(
            "workspace/list",
            ListWorkspacesRequest(type=t),
            expect=ListWorkspacesResponse,
        )


class SignedUrl:
    class Bucket(str, Enum):
        EXPORTS = "exports"
        IMPORTS = "imports"
        USER_DATA = "userData"
        PLUGIN_DATA = "pluginData"
        APP_DATA = "appData"

    class Operation(str, Enum):
        READ = "Read"
        WRITE = "Write"

    class Request(SteamshipRequest):
        bucket: SignedUrl.Bucket
        filepath: str
        operation: SignedUrl.Operation
        expires_in_minutes: int = None

    class Response(SteamshipResponse):
        bucket: str = None
        filepath: str = None
        operation: str = None
        expires_in_minutes: int = None
        signed_url: str = Field(None, alias="signedUrl")


SignedUrl.Request.update_forward_refs()
ListWorkspacesResponse.update_forward_refs()



import math

from assets.plugins.taggers.plugin_trainable_tagger import TRAINING_PARAMETERS
from steamship_tests import PLUGINS_PATH
from steamship_tests.utils.deployables import deploy_plugin
from steamship_tests.utils.fixtures import get_steamship_client

from steamship.data.plugin import HostingType
from steamship.plugin.inputs.training_parameter_plugin_input import TrainingParameterPluginInput


def test_get_training_parameters():
    """Any trainable plugin needs a Python+Lambda component that can report its trainable params.
    This steamship_tests that all the plumbing works for that to be returned"""
    client = get_steamship_client()
    tagger_path = PLUGINS_PATH / "taggers" / "plugin_trainable_tagger.py"
    # Now make a trainable tagger to train on those tags
    with deploy_plugin(
        client,
        tagger_path,
        "tagger",
        training_platform=HostingType.LAMBDA,
    ) as (tagger, tagger_version, tagger_instance):
        training_request = TrainingParameterPluginInput(plugin_instance=tagger_instance.handle)
        res = tagger_instance.get_training_parameters(training_request)
        assert res.output is not None
        params = res.output

        assert params.training_epochs is not None
        assert params.training_epochs == TRAINING_PARAMETERS.training_epochs
        assert math.isclose(
            params.testing_holdout_percent,
            TRAINING_PARAMETERS.testing_holdout_percent,
            abs_tol=0.0001,
        )
        assert params.training_params == TRAINING_PARAMETERS.training_params



from __future__ import annotations

from steamship.invocable import Invocable

# Note!
# =====
#
# This the files in this package are for Package Implementors.
# If you are using the Steamship Client, you probably are looking for either steamship.client or steamship.data
#


class PackageService(Invocable):
    """The Abstract Base Class of a Steamship Package.

    Packages may implement whatever methods they like.  To expose these methods as invocable HTTP routes,
    annotate the method with @get or @post and the route name.

    Package *implementations* are effectively stateless, though they will have stateful

    """



from assets.plugins.blockifiers.async_blockifier import ASYNC_JOB_ID, STATUS_MESSAGE
from steamship_tests import PLUGINS_PATH
from steamship_tests.utils.deployables import deploy_plugin
from steamship_tests.utils.file import upload_file
from steamship_tests.utils.fixtures import get_steamship_client


def test_e2e_async_blockifier_plugin():
    async_blockifier_plugin_path = PLUGINS_PATH / "blockifiers" / "async_blockifier.py"
    client = get_steamship_client()

    version_config_template = {
        "text_column": {"type": "string"},
        "tag_columns": {"type": "string"},
        "tag_kind": {"type": "string"},
    }
    instance_config = {"text_column": "Message", "tag_columns": "Category", "tag_kind": "Intent"}

    with deploy_plugin(
        client,
        async_blockifier_plugin_path,
        "blockifier",
        version_config_template=version_config_template,
        instance_config=instance_config,
    ) as (plugin, version, instance):
        with upload_file(client, "utterances.csv") as file:
            assert len(file.refresh().blocks) == 0
            blockify_response = file.blockify(plugin_instance=instance.handle)
            blockify_response.wait()

            # Check that the response has the correct bits set indicating that the engine handled it as an async
            # external task.
            assert blockify_response.remote_status_input
            assert blockify_response.remote_status_input.get("async_job_id") == ASYNC_JOB_ID
            assert blockify_response.remote_status_message == STATUS_MESSAGE
            assert blockify_response.retries
            assert blockify_response.retries == 1

            # Check the number of blocks
            blocks = file.refresh().blocks
            assert len(blocks) == 5
            # Check if the tags are correctly added
            for block in blocks:
                assert block.tags is not None
                assert len(block.tags) > 0
                for tag in block.tags:
                    assert tag.name is not None
                    assert tag.kind is not None



import json

import pytest
from steamship_tests import PLUGINS_PATH
from steamship_tests.utils.deployables import deploy_plugin

from steamship import MimeTypes, SteamshipError
from steamship.client import Steamship
from steamship.data.block import Block
from steamship.data.file import File
from steamship.data.tags.tag import Tag


@pytest.mark.usefixtures("client")
def test_file_upload(client: Steamship):
    a = File.create(client=client, handle="foo", content="A", mime_type=MimeTypes.MKD)
    assert a.id is not None
    assert a.mime_type == MimeTypes.MKD
    assert a.handle == "foo"

    b = File.create(client=client, content="B", mime_type=MimeTypes.TXT)
    assert b.id is not None
    assert b.mime_type == MimeTypes.TXT
    assert a.id != b.id
    assert a.raw().decode("utf-8") == "A"
    a.delete()
    b.delete()


@pytest.mark.usefixtures("client")
def test_only_blocks_or_content(client: Steamship):
    with pytest.raises(SteamshipError):
        _ = File.create(client=client, content="A", blocks=[], mime_type=MimeTypes.MKD)


@pytest.mark.usefixtures("client")
def test_file_upload_with_content_and_tags(client: Steamship):
    a = File.create(
        client=client,
        content="ABC",
        mime_type=MimeTypes.MKD,
        tags=[Tag.CreateRequest(kind="SomeKind")],
    )
    assert a.id is not None
    assert a.mime_type == MimeTypes.MKD
    assert len(a.tags) == 1
    assert a.tags[0].file_id == a.id
    assert a.tags[0].kind == "SomeKind"
    assert a.raw().decode("utf-8") == "ABC"


def test_file_import_response_dict():
    resp = File.CreateResponse(_bytes=b"some bytes", mime_type=MimeTypes.BINARY)
    to_dict = resp.dict(include={"data_", "mime_type"})
    file_create_response = File.CreateResponse.parse_obj(to_dict)
    assert resp.data_ == file_create_response.data_
    assert resp.mime_type == file_create_response.mime_type


def test_file_import_response_bytes_serialization():
    file_resp = File.CreateResponse(_bytes=b"some bytes", mime_type=MimeTypes.BINARY)
    to_dict = file_resp.dict()
    as_json_string = json.dumps(to_dict)
    as_dict_again = json.loads(as_json_string)
    assert as_dict_again == to_dict


def test_file_upload_with_blocks(client: Steamship):
    a = File.create(
        client=client,
        blocks=[
            Block.CreateRequest(text="A", tags=[Tag.CreateRequest(kind="BlockTag")]),
            Block.CreateRequest(text="B", tags=[Tag.CreateRequest(kind="BlockTag")]),
        ],
    )
    assert a.id is not None

    query_result = Block.query(client, f'file_id "{a.id}"')

    def check_blocks(block_list):
        assert len(block_list) == 2
        assert block_list[0].tags is not None
        assert len(block_list[0].tags) == 1
        assert block_list[0].tags[0].kind == "BlockTag"
        assert block_list[0].text == "A"

    assert query_result.blocks is not None
    check_blocks(query_result.blocks)

    # Let's get the file fresh
    aa = File.get(client, _id=a.id)
    check_blocks(aa.blocks)
    a.delete()


def test_file_upload_with_blocks_and_tags(client: Steamship):
    a = File.create(
        client=client,
        blocks=[
            Block.CreateRequest(text="A", tags=[Tag.CreateRequest(kind="BlockTag")]),
            Block.CreateRequest(text="B", tags=[Tag.CreateRequest(kind="BlockTag")]),
        ],
        tags=[Tag.CreateRequest(kind="FileTag")],
    )
    assert a.id is not None

    blocks = Block.query(client, f'file_id "{a.id}"')

    def check_blocks(block_list):
        assert len(block_list) == 2
        assert block_list[0].tags is not None
        assert len(block_list[0].tags) == 1
        assert block_list[0].tags[0].kind == "BlockTag"
        assert block_list[0].text == "A"

    assert blocks.blocks is not None
    check_blocks(blocks.blocks)

    # Let's get the file fresh
    aa = File.get(client, _id=a.id)
    check_blocks(aa.blocks)
    assert aa.tags is not None
    assert len(aa.tags) == 1
    assert aa.tags[0].kind == "FileTag"

    a.delete()


def test_file_upload_with_tags(client: Steamship):
    a = File.create(
        client=client,
        tags=[Tag.CreateRequest(kind="FileTag")],
    )
    assert a.id is not None

    query_result = Tag.query(client, f'filetag and file_id "{a.id}"')

    def check_tags(file_tag_list):
        assert len(file_tag_list) == 1
        assert file_tag_list[0] is not None
        assert file_tag_list[0].kind == "FileTag"

    assert query_result.tags is not None
    check_tags(query_result.tags)

    # Let's get the file fresh
    aa = File.get(client, _id=a.id)
    check_tags(aa.tags)

    a.delete()


def test_query(client: Steamship):
    a = File.create(
        client=client,
        blocks=[
            Block.CreateRequest(text="A", tags=[Tag.CreateRequest(kind="BlockTag")]),
            Block.CreateRequest(text="B"),
        ],
    )
    assert a.id is not None
    b = File.create(
        client=client,
        blocks=[Block.CreateRequest(text="A"), Block.CreateRequest(text="B")],
        tags=[Tag.CreateRequest(kind="FileTag")],
    )
    assert b.id is not None

    files = File.query(client=client, tag_filter_query='blocktag and kind "BlockTag"').files
    assert len(files) == 1
    assert files[0].id == a.id

    files = File.query(client=client, tag_filter_query='filetag and kind "FileTag"').files
    assert len(files) == 1
    assert files[0].id == b.id

    # Test serialization; This shouldn't throw
    out_file = File.get(client, _id=b.id)
    json.dumps(out_file.dict())

    a.delete()
    b.delete()


def test_file_list(client: Steamship):
    a = File.create(
        client=client,
        tags=[Tag.CreateRequest(kind="FileTag")],
    )
    b = File.create(
        client=client,
        tags=[Tag.CreateRequest(kind="FileTag")],
    )
    c = File.create(
        client=client,
        tags=[Tag.CreateRequest(kind="FileTag")],
    )

    files = File.list(client=client).files
    assert len(files) == 3
    assert a in files
    assert b in files
    assert c in files

    a.delete()
    b.delete()
    c.delete()


def test_file_refresh(client: Steamship):
    blockifier_path = PLUGINS_PATH / "blockifiers" / "blockifier.py"
    with deploy_plugin(client, blockifier_path, "blockifier") as (
        plugin,
        version,
        instance,
    ):
        file = File.create(client=client, content="This is a test.")
        assert len(file.blocks) == 0
        file.blockify(plugin_instance=instance.handle).wait()
        file.refresh()  # don't reassign file
        assert len(file.blocks) == 4
        file.delete()



from pathlib import Path

from steamship_tests import PACKAGES_PATH, PLUGINS_PATH
from steamship_tests.utils.deployables import zip_deployable


def build_asset(py_path: Path, output_file: str):
    zip_bytes = zip_deployable(py_path)
    with open(output_file, "wb") as f:
        f.write(zip_bytes)


def main():

    assets_to_build = [
        (PLUGINS_PATH / "blockifiers" / "csv_blockifier.py", "csv-blockifier.zip"),
        (PACKAGES_PATH / "demo_package.py", "demo-package.zip"),
        (PACKAGES_PATH / "configurable_hello_world.py", "configurable_hello_world.zip"),
        (PLUGINS_PATH / "blockifiers" / "blockifier.py", "dummy_blockifier.zip"),
        (PACKAGES_PATH / "bad_package.py", "bad-package.zip"),
    ]

    for path, output in assets_to_build:
        build_asset(path, output)


main()



__copyright__ = "Steamship"
__license__ = "MIT"

from typing import Any, Callable, Optional

from pydantic import BaseModel, Extra

from steamship.base.model import CamelModel


class Foo:
    def __init__(self):
        self.name = self._instance_name

    @staticmethod
    def name() -> str:
        return "static"

    def _instance_name(self) -> str:
        return "instance"


class CamelFoo(CamelModel, extra=Extra.allow):
    def __init__(self):
        super().__init__()
        self.name = self._instance_name  # ignore: F841

    @staticmethod
    def name() -> str:
        return "static"

    def _instance_name(self) -> str:
        return "instance"


class CallableFoo(CamelModel):
    name: Optional[Callable[[Any], Any]]

    def __init__(self):
        super().__init__()
        object.__setattr__(self, "name", self._instance_name)  # ignore: F841

    @staticmethod
    def name() -> str:
        return "static"

    def _instance_name(self) -> str:
        return "instance"

    def __repr_args__(self: BaseModel) -> Any:
        ret = [(key, value) for key, value in self.__dict__.items() if key != "name"]
        return ret


def test_static_instance_methods():
    assert Foo.name() == "static"
    foo = Foo()
    assert foo.name() == "instance"

    assert CamelFoo.name() == "static"
    camel_foo = CamelFoo()
    assert camel_foo.name() == "instance"

    # This is the best method! It doesn't require any.
    assert CallableFoo.name() == "static"
    callable_foo = CallableFoo()
    assert callable_foo.name() == "instance"

    # But does it infinite recurse on repr?
    assert callable_foo.__repr__()



from __future__ import annotations

from typing import Any, Dict, List, Optional

from pydantic import Field

from steamship.base.client import Client
from steamship.base.model import CamelModel
from steamship.base.request import Request
from steamship.base.response import Response


class TagQueryRequest(Request):
    tag_filter_query: str


class Tag(CamelModel):
    # Steamship client.
    client: Client = Field(None, exclude=True)

    # ID of the tag in the database.
    id: str = None

    # ID of the file associated with the tag.
    file_id: str = None

    # ID of the block associated with the tag. If not None, `start_idx` and `end_idx` should be set.
    block_id: Optional[str] = None

    # The kind of tag. See the ``TagKind`` enum class for suggestions.
    kind: str = None  # E.g. ner

    # The name of tag. See the ``DocTag``, ``TokenTag``, etc enum classes for suggestions.
    name: Optional[str] = None  # E.g. person

    # The value payload of the tag. Always a JSON-style object.
    value: Optional[Dict[str, Any]] = None

    # Character index in associated block of the start of the span of text this tag comments upon. Start-inclusive.
    start_idx: Optional[int] = None

    # Character index in associated block of the end of the span of text this tag comments upon. End-exclusive.
    end_idx: Optional[int] = None

    # The text covered by the tag.
    # Note:
    #   The text will not always be materialized into the tag object
    #   itself; you may have to fetch it with file.text[tag.start_idx:tag.end_idx]
    # Note:
    #   Changing this field will not result in changes to Steamship's database.
    #   TODO(ted): Consider refactoring as a read-only property.
    #
    text: Optional[str] = None

    class CreateRequest(Request):
        """Request to create a new Tag."""

        id: str = None
        file_id: str = None
        block_id: str = None
        kind: str = None
        name: str = None
        start_idx: int = None
        end_idx: int = None
        value: Dict[str, Any] = None

    class DeleteRequest(Request):
        id: str = None
        file_id: str = None
        block_id: str = None

    class ListRequest(Request):
        file_id: str = None
        block_id: str = None

    class ListResponse(Response):
        tags: List[Tag] = None

    @staticmethod
    def create(
        client: Client,
        file_id: str = None,
        block_id: str = None,
        kind: str = None,
        name: str = None,
        start_idx: int = None,
        end_idx: int = None,
        value: Dict[str, Any] = None,
    ) -> Tag:
        req = Tag.CreateRequest(
            file_id=file_id,
            block_id=block_id,
            kind=kind,
            name=name,
            start_idx=start_idx,
            end_idx=end_idx,
            value=value,
        )
        return client.post("tag/create", req, expect=Tag)

    def delete(self) -> Tag:
        return self.client.post(
            "tag/delete",
            Tag.DeleteRequest(id=self.id, file_id=self.file_id, block_id=self.block_id),
            expect=Tag,
        )

    def index(self, plugin_instance: Any = None):
        """Index this tag."""
        return plugin_instance.insert(self)

    @staticmethod
    def query(
        client: Client,
        tag_filter_query: str,
    ) -> TagQueryResponse:
        req = TagQueryRequest(tag_filter_query=tag_filter_query)
        res = client.post(
            "tag/query",
            payload=req,
            expect=TagQueryResponse,
        )
        return res

    # @staticmethod
    # def search(
    #     client: Client,
    #     search_index: Any,
    # ) -> TagQueryResponse:
    #     # Avoid the circular import.
    #     from steamship.data.plugin.index_plugin_instance import EmbeddingIndexPluginInstance
    #
    #     if not isinstance(search_index, EmbeddingIndexPluginInstance):
    #         raise SteamshipError(message="The search_index field")
    #
    #
    #     req = TagQueryRequest(tag_filter_query=tag_filter_query)
    #     res = client.post(
    #         "tag/query",
    #         payload=req,
    #         expect=TagQueryResponse,
    #     )
    #     return res


class TagQueryResponse(Response):
    tags: List[Tag]


Tag.ListResponse.update_forward_refs()



import pytest
from steamship_tests.utils.fixtures import get_steamship_client
from steamship_tests.utils.random import random_index, random_name

from steamship import SteamshipError, Tag

_TEST_EMBEDDER = "test-embedder"


def test_create_index():
    client = get_steamship_client()

    index_handle = random_name()
    index = client.use_plugin(
        "embedding-index",
        index_handle,
        config={"embedder": {"plugin_handle": _TEST_EMBEDDER, "fetch_if_exists": True}},
        fetch_if_exists=False,
    )

    assert index is not None

    # Duplicate creation should fail with fetch_if_exists=False
    with pytest.raises(SteamshipError):
        index = client.use_plugin(
            "embedding-index",
            index_handle,
            config={"embedder": {"plugin_handle": _TEST_EMBEDDER, "fetch_if_exists": True}},
            fetch_if_exists=False,
        )

    # Duplicate creation should fail with fetch_if_exists=False for embedder
    with pytest.raises(SteamshipError):
        index = client.use_plugin(
            "embedding-index",
            index_handle,
            config={"embedder": {"plugin_handle": _TEST_EMBEDDER, "fetch_if_exists": False}},
            fetch_if_exists=True,
        )

    index.delete()


def test_reload_and_delete_index():
    steamship = get_steamship_client()
    index = steamship.use_plugin(
        "embedding-index",
        random_name(),
        config={"embedder": {"plugin_handle": _TEST_EMBEDDER, "fetch_if_exists": True}},
        fetch_if_exists=True,
    )
    assert index.id is not None

    index2 = steamship.use_plugin(
        "embedding-index",
        index.handle,
        config={"embedder": {"plugin_handle": _TEST_EMBEDDER, "fetch_if_exists": True}},
        fetch_if_exists=True,
    )

    assert index.id == index2.id
    index.delete()

    # Having deleted it, the ID should now be different
    index3 = steamship.use_plugin(
        "embedding-index",
        index.handle,
        config={"embedder": {"plugin_handle": _TEST_EMBEDDER, "fetch_if_exists": True}},
        fetch_if_exists=True,
    )
    assert index.id != index3.id
    index3.delete()


def _list_equal(actual, expected):
    assert len(actual) == len(expected)
    assert all([a == b for a, b in zip(actual, expected)])


def test_insert_many():
    steamship = get_steamship_client()
    with random_index(steamship, _TEST_EMBEDDER) as index:
        item1 = Tag(text="Pizza", name="pizza", kind="food", value={"nums": [1, 2, 3]})
        item2 = Tag(
            text="Rocket Ship",
            name="workspace",
            kind="vehicle",
            value={"name": "Foo"},
        )

        index.insert([item1, item2])

        res = index.search(item1.text, k=100)
        res.wait()
        items = res.output.items
        assert items is not None
        assert len(items) == 2
        item0: Tag = items[0].tag
        assert item0.text == item1.text
        assert item0.name == item1.name
        assert item0.kind == item1.kind
        _list_equal(item0.value.get("nums"), item1.value.get("nums"))

        res = index.search(item2.text)
        res.wait()
        items = res.output.items
        assert items is not None
        item0: Tag = items[0].tag
        assert item0.text == item2.text
        assert item0.name == item2.name
        assert item0.kind == item2.kind
        assert item0.value.get("name") == item2.value.get("name")
        assert "_file_id" not in item0.value
        assert "_tag_id" not in item0.value
        assert "_block_id" not in item0.value


def test_duplicate_inserts():
    steamship = get_steamship_client()
    with random_index(steamship, _TEST_EMBEDDER) as index:
        # Test for suppressed re-indexing
        a1 = "Ted can eat an entire block of cheese."
        q1 = "Who can eat the most cheese"
        _ = index.insert(Tag(text=a1))
        _ = index.search(q1)


def test_index_usage():
    steamship = get_steamship_client()
    with random_index(steamship, _TEST_EMBEDDER) as index:
        a1 = "Ted can eat an entire block of cheese."
        q1 = "Who can eat the most cheese"
        _ = index.insert(Tag(text=a1))

        search_results = index.search(q1)
        search_results.wait()
        search_results = search_results.output.items
        assert len(search_results) == 1
        assert search_results[0].tag.text == a1
        assert search_results[0].tag.name is None
        assert search_results[0].tag.kind is None
        assert len(search_results[0].tag.value) == 0

        # Associate metadata
        a2 = "Armadillo shells are bulletproof."
        q2 = "What is something interesting about Armadillos?"
        a2id = "A2id"
        a2type = "A2type"
        a2metadata = {
            "id": a2id,
            "idid": f"{a2id}{a2id}",
            "boolVal": True,
            "intVal": 123,
            "floatVal": 1.2,
        }

        _ = index.insert(Tag(text=a2, name=a2id, kind=a2type, value=a2metadata))
        search_results2 = index.search(q2)
        search_results2.wait()
        search_results = search_results2.output.items
        assert len(search_results) == 1
        tag0: Tag = search_results[0].tag
        assert tag0.text == a2
        assert tag0.name == a2id
        assert tag0.kind == a2type
        assert tag0.value == a2metadata

        search_results4 = index.search(q2, k=10)
        search_results4.wait()
        search_results = search_results4.output.items
        assert len(search_results) == 2
        assert search_results[0].tag.text == a2
        assert search_results[1].tag.text == a1


def test_empty_queries():
    steamship = get_steamship_client()
    with random_index(steamship, _TEST_EMBEDDER) as index:
        a1 = "Ted can eat an entire block of cheese."
        a2 = "Joe can drink an entire glass of water."
        _ = index.insert([Tag(text=a1), Tag(text=a2)])

        with pytest.raises(SteamshipError):
            _ = index.search(None)

        with pytest.raises(SteamshipError):
            index.search("")

        with pytest.raises(SteamshipError):
            index.search("  ")



from __future__ import annotations

import json
from typing import Any, Dict, List, Optional, Type, Union

from pydantic import BaseModel, Field

from steamship.base import Task
from steamship.base.client import Client
from steamship.base.model import CamelModel
from steamship.base.request import DeleteRequest, Request
from steamship.base.response import Response
from steamship.data.search import Hit
from steamship.utils.metadata import metadata_to_str


class EmbedAndSearchRequest(Request):
    query: str
    docs: List[str]
    plugin_instance: str
    k: int = 1


class QueryResult(CamelModel):
    value: Optional[Hit] = None
    score: Optional[float] = None
    index: Optional[int] = None
    id: Optional[str] = None


class QueryResults(Request):
    items: List[QueryResult] = None


class EmbeddedItem(CamelModel):
    id: str = None
    index_id: str = None
    file_id: str = None
    block_id: str = None
    tag_id: str = None
    value: str = None
    external_id: str = None
    external_type: str = None
    metadata: Any = None
    embedding: List[float] = None

    def clone_for_insert(self) -> EmbeddedItem:
        """Produces a clone with a string representation of the metadata"""
        ret = EmbeddedItem(
            id=self.id,
            index_id=self.index_id,
            file_id=self.file_id,
            block_id=self.block_id,
            tag_id=self.tag_id,
            value=self.value,
            external_id=self.external_id,
            external_type=self.external_type,
            metadata=self.metadata,
            embedding=self.embedding,
        )
        if isinstance(ret.metadata, dict) or isinstance(ret.metadata, list):
            ret.metadata = json.dumps(ret.metadata)
        return ret


class IndexCreateRequest(Request):
    handle: str = None
    name: str = None
    plugin_instance: str = None
    fetch_if_exists: bool = True
    external_id: str = None
    external_type: str = None
    metadata: Any = None


class IndexInsertRequest(Request):
    index_id: str
    items: List[EmbeddedItem] = None
    value: str = None
    file_id: str = None
    block_type: str = None
    external_id: str = None
    external_type: str = None
    metadata: Any = None
    reindex: bool = True


class IndexItemId(CamelModel):
    index_id: str = None
    id: str = None


class IndexInsertResponse(Response):
    item_ids: List[IndexItemId] = None


class IndexEmbedRequest(Request):
    id: str


class IndexEmbedResponse(Response):
    id: Optional[str] = None


class IndexSearchRequest(Request):
    id: str
    query: str = None
    queries: List[str] = None
    k: int = 1
    include_metadata: bool = False


class IndexSnapshotRequest(Request):
    index_id: str
    window_size: int = None  # Used for unit testing


class IndexSnapshotResponse(Response):
    id: Optional[str] = None
    snapshot_id: str


class ListSnapshotsRequest(Request):
    id: str = None


class ListSnapshotsResponse(Response):
    snapshots: List[IndexSnapshotResponse]


class ListItemsRequest(Request):
    id: str = None
    file_id: str = None
    block_id: str = None
    span_id: str = None


class ListItemsResponse(Response):
    items: List[EmbeddedItem]


class DeleteSnapshotsRequest(Request):
    snapshot_id: str = None


class DeleteSnapshotsResponse(Response):
    snapshot_id: str = None


class EmbeddingIndex(CamelModel):
    """A persistent, read-optimized index over embeddings."""

    client: Client = Field(None, exclude=True)
    id: str = None
    handle: str = None
    name: str = None
    plugin: str = None
    external_id: str = None
    external_type: str = None
    metadata: str = None

    @classmethod
    def parse_obj(cls: Type[BaseModel], obj: Any) -> BaseModel:
        # TODO (enias): This needs to be solved at the engine side
        if "embeddingIndex" in obj:
            obj = obj["embeddingIndex"]
        elif "index" in obj:
            obj = obj["index"]
        return super().parse_obj(obj)

    def insert_file(
        self,
        file_id: str,
        block_type: str = None,
        external_id: str = None,
        external_type: str = None,
        metadata: Union[int, float, bool, str, List, Dict] = None,
        reindex: bool = True,
    ) -> IndexInsertResponse:
        if isinstance(metadata, dict) or isinstance(metadata, list):
            metadata = json.dumps(metadata)

        req = IndexInsertRequest(
            index_id=self.id,
            file_id=file_id,
            blockType=block_type,
            external_id=external_id,
            external_type=external_type,
            metadata=metadata,
            reindex=reindex,
        )
        return self.client.post(
            "embedding-index/item/create",
            req,
            expect=IndexInsertResponse,
        )

    def insert_many(
        self,
        items: List[Union[EmbeddedItem, str]],
        reindex: bool = True,
    ) -> IndexInsertResponse:
        new_items = []
        for item in items:
            if isinstance(item, str):
                new_items.append(EmbeddedItem(value=item))
            else:
                new_items.append(item)

        req = IndexInsertRequest(
            index_id=self.id,
            items=[item.clone_for_insert() for item in new_items],
            reindex=reindex,
        )
        return self.client.post(
            "embedding-index/item/create",
            req,
            expect=IndexInsertResponse,
        )

    def insert(
        self,
        value: str,
        external_id: str = None,
        external_type: str = None,
        metadata: Union[int, float, bool, str, List, Dict] = None,
        reindex: bool = True,
    ) -> IndexInsertResponse:

        req = IndexInsertRequest(
            index_id=self.id,
            value=value,
            external_id=external_id,
            external_type=external_type,
            metadata=metadata_to_str(metadata),
            reindex=reindex,
        )
        return self.client.post(
            "embedding-index/item/create",
            req,
            expect=IndexInsertResponse,
        )

    def embed(
        self,
    ) -> Task[IndexEmbedResponse]:
        req = IndexEmbedRequest(id=self.id)
        return self.client.post(
            "embedding-index/embed",
            req,
            expect=IndexEmbedResponse,
        )

    def create_snapshot(self) -> IndexSnapshotResponse:
        req = IndexSnapshotRequest(index_id=self.id)
        return self.client.post(
            "embedding-index/snapshot/create",
            req,
            expect=IndexSnapshotResponse,
        )

    def list_snapshots(self) -> ListSnapshotsResponse:
        req = ListSnapshotsRequest(id=self.id)
        return self.client.post(
            "embedding-index/snapshot/list",
            req,
            expect=ListSnapshotsResponse,
        )

    def list_items(
        self,
        file_id: str = None,
        block_id: str = None,
        span_id: str = None,
    ) -> ListItemsResponse:
        req = ListItemsRequest(id=self.id, file_id=file_id, block_id=block_id, spanId=span_id)
        return self.client.post(
            "embedding-index/item/list",
            req,
            expect=ListItemsResponse,
        )

    def delete_snapshot(
        self,
        snapshot_id: str,
    ) -> DeleteSnapshotsResponse:
        req = DeleteSnapshotsRequest(snapshotId=snapshot_id)
        return self.client.post(
            "embedding-index/snapshot/delete",
            req,
            expect=DeleteSnapshotsResponse,
        )

    def delete(self) -> EmbeddingIndex:
        return self.client.post(
            "embedding-index/delete",
            DeleteRequest(id=self.id),
            expect=EmbeddingIndex,
        )

    def search(
        self,
        query: Union[str, List[str]],
        k: int = 1,
        include_metadata: bool = False,
    ) -> Task[QueryResults]:
        if isinstance(query, list):
            req = IndexSearchRequest(
                id=self.id, queries=query, k=k, include_metadata=include_metadata
            )
        else:
            req = IndexSearchRequest(
                id=self.id, query=query, k=k, include_metadata=include_metadata
            )
        ret = self.client.post(
            "embedding-index/search",
            req,
            expect=QueryResults,
        )

        return ret

    @staticmethod
    def create(
        client: Client,
        handle: str = None,
        name: str = None,
        embedder_plugin_instance_handle: str = None,
        fetch_if_exists: bool = True,
        external_id: str = None,
        external_type: str = None,
        metadata: Any = None,
    ) -> EmbeddingIndex:
        req = IndexCreateRequest(
            handle=handle,
            name=name,
            plugin_instance=embedder_plugin_instance_handle,
            fetch_if_exists=fetch_if_exists,
            external_id=external_id,
            external_type=external_type,
            metadata=metadata,
        )
        return client.post(
            "embedding-index/create",
            req,
            expect=EmbeddingIndex,
        )



from __future__ import annotations

import io
import json
import logging
from typing import Any, Dict, Generic, Optional, TypeVar, Union

from pydantic import BaseModel
from pydantic.generics import GenericModel

from steamship.base import MimeTypes, SteamshipError, Task, TaskState
from steamship.base.client import Client
from steamship.base.error import DEFAULT_ERROR_MESSAGE
from steamship.base.mime_types import ContentEncodings
from steamship.base.model import CamelModel

# If this isn't present, Localstack won't show logs
from steamship.utils.binary_utils import flexi_create

logging.getLogger().setLevel(logging.INFO)


class Http(CamelModel):
    status: int = None
    # If true, we're signaling to the Steamship Proxy that the `data` field of the SteamshipResponse object
    # has been wrapped in base64. In this situation, we can return the bytes within directly to the Proxy
    # caller without interpreting it.
    base64_wrapped: bool = None
    headers: Dict[str, str] = None


T = TypeVar("T")


class InvocableResponse(GenericModel, Generic[T]):
    """Mirrors the Response object in the Steamship server."""

    data: T = None  # Data for successful or synchronous requests.
    status: Task = None  # Reporting for errors and async status
    http: Http = None  # Additional HTTP information for Steamship Proxy (headers, etc)

    def __init__(
        self,
        status: Task = None,
        error: SteamshipError = None,
        http: Http = None,
        data: Any = None,
        string: str = None,
        json: Any = None,
        _bytes: Union[bytes, io.BytesIO] = None,
        mime_type=None,
    ):
        super().__init__()
        # Note:
        # This function has to be very defensively coded since Any errors thrown here will not be returned
        # to the end-user via our proxy (as this is the constructor for the response itself!)
        if http is not None:
            self.http = http
        else:
            self.http = Http(status=200, headers={})

        try:
            self.set_data(data=data, string=string, json=json, _bytes=_bytes, mime_type=mime_type)
        except Exception as ex:
            logging.error("Exception within Response.__init__.", exc_info=ex)
            if error is not None:
                if error.message:
                    error.message = f"{error.message}. Also found error - unable to serialize data to response. {ex}"
                else:
                    error.message = f"Unable to serialize data to response. {ex}"
            else:
                error = SteamshipError(message=f"Unable to serialize data to response. {ex}")
            logging.error(error, exc_info=error)

        # Handle the task provided
        if status is None:
            self.status = Task()
        elif isinstance(status, Task):
            self.status = status
        else:
            self.status = Task()
            self.status.state = TaskState.failed
            self.status.status_message = (
                f"Status field of response should be of type Task. "
                f"Instead was of type {type(status)} and had value {status}."
            )

        if error:
            self.status.state = TaskState.failed
            self.status.status_message = error.message
            self.status.status_suggestion = error.suggestion
            self.status.status_code = error.code
            logging.error(
                "steamship.invocable.response - Response created with error.", exc_info=error
            )
        else:
            if self.status.state is None:
                self.status.state = TaskState.succeeded

    def set_data(
        self,
        data: Any = None,
        string: str = None,
        json: Any = None,
        _bytes: Union[bytes, io.BytesIO] = None,
        mime_type=None,
    ):
        data, mime_type, encoding = flexi_create(
            data=data, string=string, json=json, _bytes=_bytes, mime_type=mime_type
        )

        self.data = data

        self.http.headers = self.http.headers or {}
        self.http.headers["Content-Type"] = mime_type or MimeTypes.BINARY

        if encoding == ContentEncodings.BASE64:
            self.http.base64_wrapped = True

    @staticmethod
    def error(
        code: int,
        message: Optional[str] = None,
        error: Optional[SteamshipError] = None,
        exception: Optional[Exception] = None,
        prefix: Optional[str] = None,
    ) -> InvocableResponse[T]:
        """Merges a number of error channels into one unified Response object.

        Aggregates all possible messages into a single " | "-delimeted error message.

        If the final resulting error message is non-null, prefixes with the provided `prefix`
        """
        # Use or create the return error
        error = error or SteamshipError()

        messages = []
        if error.message != DEFAULT_ERROR_MESSAGE:
            messages.append(error.message)

        # Set or append the additional message
        if message is not None and message not in messages:
            messages.append(message)

        # Set or append the exception
        if exception is not None:
            exception_str = f"{exception}"
            if exception_str not in messages:
                messages.append(exception_str)

        messages = [m.strip() for m in messages if m is not None and len(m.strip())]
        if len(messages) > 0:
            error.message = " | ".join(messages)

        # Finally, add the prefix if requested.
        if prefix and error.message:
            error.message = f"{prefix}{error.message}"

        return InvocableResponse(error=error, http=Http(status=code))

    @staticmethod
    def from_obj(obj: Any) -> InvocableResponse:  # noqa: C901
        if obj is None:
            return InvocableResponse.error(500, "Handler provided no response.")

        if isinstance(obj, InvocableResponse):
            return obj
        elif isinstance(obj, SteamshipError):
            return InvocableResponse.error(500, error=obj)
        elif isinstance(obj, Exception):
            return InvocableResponse.error(500, error=SteamshipError(error=obj))
        elif isinstance(obj, io.BytesIO):
            return InvocableResponse(_bytes=obj)
        elif isinstance(obj, dict):
            return InvocableResponse(json=obj)
        elif isinstance(obj, str):
            return InvocableResponse(string=obj)
        elif isinstance(obj, (float, int, bool)):
            return InvocableResponse(json=obj)
        elif isinstance(obj, CamelModel):
            return InvocableResponse(json=obj.dict(by_alias=True))
        elif isinstance(obj, BaseModel):
            return InvocableResponse(json=obj.dict())

        return InvocableResponse.error(
            500, message=f"Handler provided unknown response type: {type(obj)}"
        )

    def post_update(self, client: Client):
        """Pushes this response object to the corresponding Task on the Steamship Engine.

        Typically apps and plugins return their results to the Engine synchronously via HTTP.
        But sometimes that's not practice -- for example:

        - Microsoft's OCR endpoint returns a Job Token that can be exchanged for updates, and eventually a result.
        - Google's AutoML can take 20-30 minutes to train.
        - Fine-tuning BERT on ECS can take an arbitrarily long amount of time.

        In these cases, it can be useful for the package/plugin to occasionally post updates to the Engine outside
        of the Engine's initial synchronous request-response conversation.
        """
        if self.status is None or self.status.task_id is None:
            raise SteamshipError(
                message="An App/Plugin response can only be pushed to the Steamship Engine if "
                + "it is associated with a Task. Please set the `status.task_id` field."
            )
        if client is None:
            raise SteamshipError(
                message="Unable to push Response to Steamship: Associated client is None"
            )

        # Create a task object
        task = Task(client=client, task_id=self.status.task_id)
        update_fields = set()

        if self.status.state is not None:
            task.state = self.status.state
            update_fields.add("state")

        if self.status.status_message is not None:
            task.status_message = self.status.status_message
            update_fields.add("status_message")

        if self.status.status_suggestion is not None:
            task.status_suggestion = self.status.status_suggestion
            update_fields.add("status_suggestion")

        if self.data is not None:
            # This object itself should always be the output of the Training Task object.
            task.output = json.dumps(self.data)
            update_fields.add("output")

        task.post_update(fields=update_fields)



import logging

from assets.packages.signed_url_package import TEST_STRING
from steamship_tests import PACKAGES_PATH
from steamship_tests.utils.deployables import deploy_package
from steamship_tests.utils.fixtures import get_steamship_client

from steamship.data.workspace import SignedUrl
from steamship.utils.signed_urls import upload_to_signed_url


def test_signed_url_package():

    client = get_steamship_client()
    demo_package_path = PACKAGES_PATH / "signed_url_package.py"

    with deploy_package(client, demo_package_path) as (_, _, instance):
        response = instance.invoke("writeReadSignedURL")
        assert response is not None
        assert response == TEST_STRING


def test_signed_url_package_read_only():
    client = get_steamship_client()
    workspace = client.get_workspace()
    filepath = "testPath"
    write_signed_url = workspace.create_signed_url(
        SignedUrl.Request(
            bucket=SignedUrl.Bucket.PLUGIN_DATA,
            filepath=filepath,
            operation=SignedUrl.Operation.WRITE,
        )
    ).signed_url

    logging.info(f"Got signed url for writing: {write_signed_url}")

    upload_to_signed_url(write_signed_url, bytes(TEST_STRING, "utf-8"))
    logging.info("Uploaded to signed url successfully.")

    demo_package_path = PACKAGES_PATH / "signed_url_package.py"
    with deploy_package(client, demo_package_path) as (_, _, instance):
        response = instance.invoke("readSignedURL", filepath=filepath)
        assert response is not None
        assert response == TEST_STRING



"""This class is a helper for plugins to use models hosted on Hugging Face.

It uses asyncio parallelism to make many http requests simultaneously.
"""

import asyncio
import logging
import time
from http import HTTPStatus
from typing import List, Optional

import aiohttp
from aiohttp import ClientTimeout

from steamship import Block, SteamshipError


async def _model_call(
    session, text: str, api_url, headers, additional_params: dict = None, use_gpu: bool = False
) -> Optional[list]:
    additional_params = additional_params or {}
    json_input = {
        "inputs": text or "",
        "parameters": additional_params,
        "options": {"use_gpu": use_gpu, "wait_for_model": False},
    }
    ok_response, nok_response = None, None

    max_error_retries = 3

    """
    Hugging Face returns an error that says that the model is currently loading
    if it believes you have 'too many' requests simultaneously, so the logic retries in this case, but fails on
    other errors.
    """
    tries = 0
    while tries <= max_error_retries:
        async with session.post(api_url, headers=headers, json=json_input) as response:
            if response.status == HTTPStatus.OK and response.content_type == "application/json":
                ok_response = await response.json()
                logging.info(ok_response)
                return ok_response
            else:
                nok_response = await response.text()
                if "is currently loading" not in nok_response:
                    logging.info(
                        f'Received text response "{nok_response}" for input text "{text}" [attempt {tries}/{max_error_retries}]'
                    )
                    tries += 1
                else:
                    await asyncio.sleep(1)
    if ok_response is None:
        raise SteamshipError(
            message="Unable to query Hugging Face model",
            internal_message=f"HF returned error: {nok_response} after {tries} attempts",
        )
    return ok_response


async def _model_calls(
    texts: List[str],
    api_url: str,
    headers,
    timeout_seconds: int,
    additional_params: dict = None,
    use_gpu: bool = False,
) -> List[list]:
    async with aiohttp.ClientSession(timeout=ClientTimeout(total=timeout_seconds)) as session:
        tasks = []
        for text in texts:
            tasks.append(
                asyncio.ensure_future(
                    _model_call(
                        session,
                        text,
                        api_url,
                        headers=headers,
                        additional_params=additional_params,
                        use_gpu=use_gpu,
                    )
                )
            )

        return await asyncio.gather(*tasks)


def get_huggingface_results(
    blocks: List[Block],
    hf_model_path: str,
    hf_bearer_token: str,
    additional_params: dict = None,
    timeout_seconds: int = 30,
    use_gpu: bool = False,
) -> List[list]:
    api_url = f"https://api-inference.huggingface.co/models/{hf_model_path}"
    headers = {"Authorization": f"Bearer {hf_bearer_token}"}
    start_time = time.perf_counter()
    results = asyncio.run(
        _model_calls(
            [block.text for block in blocks],
            api_url,
            headers,
            timeout_seconds=timeout_seconds,
            additional_params=additional_params,
            use_gpu=use_gpu,
        )
    )
    total_time = time.perf_counter() - start_time
    logging.info(
        f"Completed {len(blocks)} blocks in {total_time} seconds. ({float(len(blocks)) / total_time} bps)"
    )
    return results



from .config import Config
from .invocable import Invocable, get, post
from .invocable_request import InvocableRequest, Invocation, InvocationContext, LoggingConfig
from .invocable_response import InvocableResponse
from .lambda_handler import create_handler, safe_handler
from .package_service import PackageService

__all__ = [
    "Invocable",
    "create_handler",
    "Config",
    "Invocation",
    "InvocableRequest",
    "InvocableResponse",
    "get",
    "post",
    "InvocationContext",
    "LoggingConfig",
    "PackageService",
    "safe_handler",
]



import pytest

from steamship.client import Steamship
from steamship.data.block import Block
from steamship.data.file import File
from steamship.data.tags.tag import Tag


@pytest.mark.usefixtures("client")
def test_query(client: Steamship):
    a = File.create(
        client=client,
        blocks=[
            Block.CreateRequest(text="A", tags=[Tag.CreateRequest(kind="BlockTag")]),
            Block.CreateRequest(text="B"),
        ],
    )
    assert a.id is not None
    a = a.refresh()
    b = File.create(
        client=client,
        blocks=[
            Block.CreateRequest(text="A"),
            Block.CreateRequest(text="B", tags=[Tag.CreateRequest(kind="Test")]),
        ],
        tags=[Tag.CreateRequest(kind="FileTag")],
    )
    assert b.id is not None
    b = b.refresh()

    blocks = Block.query(client=client, tag_filter_query='blocktag and kind "BlockTag"').blocks
    assert len(blocks) == 1
    assert blocks[0].id == a.blocks[0].id

    blocks = Block.query(client=client, tag_filter_query='blocktag and kind "Test"').blocks
    assert len(blocks) == 1
    assert blocks[0].id == b.blocks[1].id

    a.delete()
    b.delete()



from __future__ import annotations

from steamship.base.model import CamelModel


class ExportPluginInput(CamelModel):
    plugin_instance: str = None
    id: str = None
    handle: str = None
    type: str = None
    filename: str = None
    query: str = None



from .tag import Tag
from .tag_constants import (
    DocTag,
    EmotionTag,
    EntityTag,
    GenerationTag,
    IntentTag,
    SentimentTag,
    TagKind,
    TagValue,
    TokenTag,
)

__all__ = [
    "DocTag",
    "EmotionTag",
    "SentimentTag",
    "IntentTag",
    "EntityTag",
    "Tag",
    "TagKind",
    "TokenTag",
    "TagValue",
    "GenerationTag",
]



import base64
import io
import json as jsonlib
import logging
from typing import Any, Tuple, Union

from pydantic import BaseModel

from steamship.base import MimeTypes, SteamshipError
from steamship.base.mime_types import ContentEncodings
from steamship.base.model import CamelModel


def guess_mime(obj: Any, provided_mime: str = None) -> str:
    if provided_mime is not None:
        return provided_mime
    if isinstance(obj, (str, int, float, bool)):
        return MimeTypes.TXT
    return MimeTypes.BINARY


def to_b64(obj: Any) -> str:
    ret_bytes = obj
    if isinstance(obj, bytes):
        ret_bytes = obj
    elif isinstance(obj, str):
        ret_bytes = ret_bytes.encode("utf-8")
    else:
        ret_bytes = str(obj).encode("utf-8")
    return base64.b64encode(ret_bytes).decode("utf-8")


def flexi_create(  # noqa: C901
    base64string: str = None,
    data: Any = None,
    string: str = None,
    json: Any = None,
    _bytes: Union[bytes, io.BytesIO] = None,
    mime_type=None,
    force_base64=False,
) -> Tuple[Any, Union[None, str], Union[None, str]]:  # TODO (Enias): Review
    """
    It's convenient for some constructors to accept a variety of input types:
    - data (your choice)
    - string
    - json
    - bytes

    .. And have them all homogenized.
    """

    try:
        if base64string is not None:
            return base64string, mime_type or MimeTypes.BINARY, ContentEncodings.BASE64

        ret_data = None  # the body of the result
        ret_mime = None  # for the Content-Type field
        ret_encoding = None  # for the Content-Encoding field
        is_b64 = False

        if data is not None:
            ret_data, ret_mime = data, mime_type or guess_mime(data, mime_type)

        elif string is not None:
            ret_data, ret_mime = string, mime_type or MimeTypes.TXT

        elif json is not None:
            ret_mime = mime_type or MimeTypes.JSON

            if isinstance(json, CamelModel):
                ret_dict = json.dict(by_alias=True)
                ret_data = ret_dict
            elif isinstance(json, BaseModel):
                ret_dict = json.dict()
                ret_data = ret_dict
            else:
                ret_data = json

        elif _bytes is not None:
            if isinstance(_bytes, io.BytesIO):
                _bytes = _bytes.getvalue()  # Turn it into regular bytes
            ret_data, ret_mime = (
                base64.b64encode(_bytes).decode("utf-8"),
                mime_type or ret_mime or MimeTypes.BINARY,
            )
            is_b64 = True
            ret_encoding = ContentEncodings.BASE64

        if ret_data is not None:
            if force_base64 is False:
                return ret_data, ret_mime, ret_encoding
            if is_b64 is True:
                return ret_data, ret_mime, ContentEncodings.BASE64
            else:
                if json is not None or (data is not None and ret_mime == MimeTypes.JSON):
                    # If it was JSON, we need to dump the object first!
                    # Otherwise it will end up getting turned to the Python's object representation format
                    # which will result in invalid JSON
                    ret_data = jsonlib.dumps(ret_data)

                return (
                    to_b64(ret_data),
                    ret_mime or MimeTypes.BINARY,
                    ContentEncodings.BASE64,
                )

        return None, None, None
    except Exception as ex:
        logging.error("Exception thrown trying to encode data", exc_info=ex)
        raise SteamshipError(
            message="There was an exception thrown while trying to encode your package/plugin data.",
            error=ex,
        )



# TODO: It should fail if the docs field is empty.
# TODO: It should fail if the file hasn't been converted.
from enum import Enum

from steamship_tests.utils.fixtures import get_steamship_client

from steamship import Block, DocTag, File, MimeTypes, PluginInstance, Steamship
from steamship.base import TaskState


def count_blocks_with_tag(blocks: [Block], tag_kind: Enum, tag_name: Enum):
    c = 0
    for block in blocks:
        if any([tag.kind == tag_kind.value and tag.name == tag_name.value for tag in block.tags]):
            c += 1
    return c


def count_tags(blocks: [Block], tag_kind: Enum, tag_name: Enum):
    c = 0
    for block in blocks:
        tag_matches = [
            1 if tag.kind == tag_kind.value and tag.name == tag_name.value else 0
            for tag in block.tags
        ]
        c += sum(tag_matches)
    return c


def tag_file(client: Steamship, parser_instance_handle: str):
    t = "A Poem"
    p1_1 = "Roses are red."
    p1_2 = "Violets are blue."
    p2_1 = "Sugar is sweet, and I love you."

    content = f"# {t}\n\n{p1_1} {p1_2}\n\n{p2_1}"

    a = File.create(
        client,
        content=content,
        mime_type=MimeTypes.MKD,
    )
    assert a.id is not None
    assert a.mime_type == MimeTypes.MKD

    a.blockify(plugin_instance="markdown-blockifier-default-1.0").wait()

    raw = a.raw()
    assert raw.decode("utf-8") == content

    # The following steamship_tests should be updated once the Tag query basics are merged.
    # Instead of querying and filtering, do a query with a tag filter
    q1 = a.refresh()
    assert count_blocks_with_tag(q1.blocks, DocTag.DOCUMENT, DocTag.H1) == 1
    assert q1.blocks[0].text == t

    # Instead of re-filtering previous result, do a new tag filter query
    assert count_blocks_with_tag(q1.blocks, DocTag.DOCUMENT, DocTag.PARAGRAPH) == 2
    assert q1.blocks[1].text == f"{p1_1} {p1_2}"

    # The sentences aren't yet parsed out!
    # Instead of re-filtering again, do a new tag filter query
    assert count_blocks_with_tag(q1.blocks, DocTag.DOCUMENT, DocTag.SENTENCE) == 0

    # Now we parse
    task = a.tag(plugin_instance=parser_instance_handle)
    assert task is not None
    assert task.state == TaskState.waiting

    task.wait()
    assert task is not None
    assert task.state == TaskState.succeeded

    # Now the sentences should be parsed!
    # Again, should rewrite these once tag queries are integrated
    q2 = a.refresh()
    assert count_tags(q2.blocks, DocTag.DOCUMENT, DocTag.SENTENCE) == 4

    a.delete()


def test_parse_file():
    steamship = get_steamship_client()
    parser = PluginInstance.create(steamship, plugin_handle="test-tagger")
    tag_file(steamship, parser.handle)



from steamship_tests import PLUGINS_PATH
from steamship_tests.utils.deployables import deploy_plugin
from steamship_tests.utils.fixtures import get_steamship_client

from steamship import PluginInstance


def test_e2e_parser():
    client = get_steamship_client()
    tagger_plugin_path = PLUGINS_PATH / "taggers" / "plugin_configurable_tagger.py"
    config_template = {
        "tagKind": {"type": "string"},
        "tagName": {"type": "string"},
        "numberValue": {"type": "number"},
        "booleanValue": {"type": "boolean"},
    }
    instance_config1 = {
        "tagKind": "testTagKind",
        "tagName": "testTagName",
        "numberValue": 3,
        "booleanValue": True,
    }

    with deploy_plugin(
        client,
        tagger_plugin_path,
        "tagger",
        version_config_template=config_template,
        instance_config=instance_config1,
    ) as (plugin, version, instance):
        test_doc = "Hi there"
        res = instance.tag(doc=test_doc)
        res.wait()
        assert res.output is not None
        assert len(res.output.file.blocks) == 1
        assert res.output.file.blocks[0].text == test_doc

        # Validate configured content
        assert len(res.output.file.tags) == 1
        tag = res.output.file.tags[0]
        assert tag.name == instance_config1["tagName"]
        assert tag.kind == instance_config1["tagKind"]
        tag_value = tag.value
        assert tag_value["numberValue"] == instance_config1["numberValue"]
        assert tag_value["booleanValue"] == instance_config1["booleanValue"]

        instance_config2 = {
            "tagKind": "testTagKind2",
            "tagName": "testTagName2",
            "numberValue": 4,
            "booleanValue": False,
        }

        instance2 = PluginInstance.create(
            client,
            plugin_id=plugin.id,
            plugin_version_id=version.id,
            config=instance_config2,
        )
        assert instance2 is not None

        res = instance2.tag(doc=test_doc)
        res.wait()
        assert res.output is not None
        assert len(res.output.file.blocks) == 1
        assert res.output.file.blocks[0].text == test_doc

        # Validate configured content
        assert len(res.output.file.tags) == 1
        tag = res.output.file.tags[0]
        assert tag.name == instance_config2["tagName"]
        assert tag.kind == instance_config2["tagKind"]
        tag_value = tag.value
        assert tag_value["numberValue"] == instance_config2["numberValue"]
        assert tag_value["booleanValue"] == instance_config2["booleanValue"]



from .block import Block
from .embeddings import EmbeddingIndex
from .file import File
from .package import Package, PackageInstance, PackageVersion
from .plugin import Plugin, PluginInstance, PluginVersion
from .tags import (
    DocTag,
    EmotionTag,
    EntityTag,
    GenerationTag,
    IntentTag,
    SentimentTag,
    Tag,
    TagKind,
    TagValue,
    TokenTag,
)
from .workspace import Workspace

__all__ = [
    "Package",
    "PackageInstance",
    "PackageVersion",
    "Block",
    "EmbeddingIndex",
    "EntityTag",
    "File",
    "GenerationTag",
    "IntentTag",
    "Plugin",
    "PluginInstance",
    "PluginVersion",
    "Workspace",
    "DocTag",
    "EmotionTag",
    "SentimentTag",
    "Tag",
    "TagKind",
    "TokenTag",
    "TagValue",
]



"""When a Plugin's operation is tied to a task, the plugin can update that task while it is still running,
or asynchronously at any time."""

import pytest
from steamship_tests.plugin.unit.trainable.util import create_dummy_training_task
from steamship_tests.utils.fixtures import get_steamship_client

from steamship import SteamshipError
from steamship.base.tasks import Task, TaskState
from steamship.invocable import InvocableResponse


def test_response_post_update_fails_when_no_task_present():
    client = get_steamship_client()
    response = InvocableResponse()
    with pytest.raises(SteamshipError):
        response.post_update(client)

    # No task_id
    response2 = InvocableResponse(status=Task())
    with pytest.raises(SteamshipError):
        response2.post_update(client)


def test_response_post_update_can_update_task():
    client = get_steamship_client()
    task = create_dummy_training_task(client)

    new_state = TaskState.failed
    new_message = "HI THERE"
    new_output = {"a": 3}

    assert task.state != new_state
    assert task.status_message != new_message
    assert task.output != new_output

    response = InvocableResponse(status=task)

    response.status.state = new_state
    response.status.status_message = new_message
    response.status.output = new_output

    # Sanity check: we'll prove that caling task.check() resets this..
    task.refresh()

    # Assert not equal
    assert task.state != new_state
    assert task.status_message != new_message
    assert task.output != new_output

    # And override again
    response.status.state = TaskState.failed
    response.status.status_message = new_message
    response.set_data(json=new_output)

    # Now we call post_update
    response.post_update(client)

    # Call task.check
    with pytest.raises(SteamshipError):
        task.refresh()



from __future__ import annotations

import io
import logging
from enum import Enum
from typing import TYPE_CHECKING, Any, List, Optional, Type, Union

from pydantic import BaseModel, Field

from steamship import SteamshipError
from steamship.base.client import Client
from steamship.base.model import CamelModel
from steamship.base.request import GetRequest, IdentifierRequest, Request
from steamship.base.response import Response
from steamship.base.tasks import Task
from steamship.data.block import Block
from steamship.data.embeddings import EmbeddingIndex
from steamship.data.tags import Tag
from steamship.utils.binary_utils import flexi_create

if TYPE_CHECKING:
    from steamship.data.operations.tagger import TagResponse


class FileUploadType(str, Enum):
    FILE = "file"  # The CreateRequest contains a file upload that should be used
    VALUE = "value"  # The Create Request contains a `text` field that should be used
    FILE_IMPORTER = (
        "fileImporter"  # The CreateRequest contains a fileImporter handle that should be used
    )
    BLOCKS = "blocks"  # The CreateRequest contains blocks and tags that should be read in directly


_logger = logging.getLogger(__name__)


class FileClearResponse(Response):
    id: str


class ListFileRequest(Request):
    pass


class ListFileResponse(Response):
    files: List[File]


class FileQueryRequest(Request):
    tag_filter_query: str


class File(CamelModel):
    """A file."""

    client: Client = Field(None, exclude=True)
    id: str = None
    handle: str = None
    mime_type: str = None
    workspace_id: str = None
    blocks: List[Block] = []
    tags: List[Tag] = []
    filename: str = None

    class CreateRequest(Request):
        value: str = None
        data: str = None
        id: str = None
        url: str = None
        handle: str = None
        filename: str = None
        type: FileUploadType = None
        mime_type: str = None
        blocks: Optional[List[Block.CreateRequest]] = []
        tags: Optional[List[Tag.CreateRequest]] = []
        plugin_instance: str = None

    class CreateResponse(Response):
        data_: Any = None
        mime_type: str = None

        def __init__(
            self,
            data: Any = None,
            string: str = None,
            _bytes: Union[bytes, io.BytesIO] = None,
            json: io.BytesIO = None,
            mime_type: str = None,
        ):
            super().__init__()
            data, mime_type, encoding = flexi_create(
                data=data, string=string, json=json, _bytes=_bytes, mime_type=mime_type
            )
            self.data_ = data
            self.mime_type = mime_type

        @classmethod
        def parse_obj(cls: Type[BaseModel], obj: Any) -> Response:
            obj["data"] = obj.get("data") or obj.get("data_")
            if "data_" in obj:
                del obj["data_"]
            return super().parse_obj(obj)

    @classmethod
    def parse_obj(cls: Type[BaseModel], obj: Any) -> BaseModel:
        # TODO (enias): This needs to be solved at the engine side
        obj = obj["file"] if "file" in obj else obj
        return super().parse_obj(obj)

    def delete(self) -> File:
        return self.client.post(
            "file/delete",
            IdentifierRequest(id=self.id),
            expect=File,
        )

    @staticmethod
    def get(
        client: Client,
        _id: str = None,
        handle: str = None,
    ) -> File:
        return client.post(
            "file/get",
            IdentifierRequest(id=_id, handle=handle),
            expect=File,
        )

    @staticmethod
    def create(
        client: Client,
        content: Union[str, bytes] = None,
        mime_type: str = None,
        handle: str = None,
        blocks: List[Block.CreateRequest] = None,
        tags: List[Tag.CreateRequest] = None,
    ) -> File:

        if content is None and blocks is None:
            if tags is None:
                raise SteamshipError(message="Either filename, content, or tags must be provided.")
            else:
                content = ""
        if content is not None and blocks is not None:
            raise SteamshipError(
                message="Please provide only `blocks` or `content` to `File.create`."
            )

        if blocks is not None:
            upload_type = FileUploadType.BLOCKS
        elif content is not None:
            upload_type = FileUploadType.FILE
        else:
            raise Exception("Unable to determine upload type.")

        req = File.CreateRequest(
            handle=handle,
            type=upload_type,
            mime_type=mime_type,
            blocks=blocks,
            tags=tags,
        )

        file_data = (
            ("file-part", content, "multipart/form-data")
            if upload_type != FileUploadType.BLOCKS
            else None
        )

        # Defaulting this here, as opposed to in the Engine, because it is processed by Vapor
        return client.post(
            "file/create",
            payload=req,
            file=file_data,
            expect=File,
        )

    @staticmethod
    def create_with_plugin(
        client: Client,
        plugin_instance: str,
        url: str = None,
        mime_type: str = None,
    ) -> Task[File]:

        req = File.CreateRequest(
            type=FileUploadType.FILE_IMPORTER,
            url=url,
            mime_type=mime_type,
            plugin_instance=plugin_instance,
        )

        return client.post("file/create", payload=req, expect=File, as_background_task=True)

    def refresh(self) -> File:
        refreshed = File.get(self.client, self.id)
        self.__init__(**refreshed.dict())
        self.client = refreshed.client
        return self

    @staticmethod
    def query(
        client: Client,
        tag_filter_query: str,
    ) -> FileQueryResponse:

        req = FileQueryRequest(tag_filter_query=tag_filter_query)
        res = client.post(
            "file/query",
            payload=req,
            expect=FileQueryResponse,
        )
        return res

    def raw(self):
        return self.client.post(
            "file/raw",
            payload=GetRequest(
                id=self.id,
            ),
            raw_response=True,
        )

    def blockify(self, plugin_instance: str = None, wait_on_tasks: List[Task] = None) -> Task:
        from steamship.data.operations.blockifier import BlockifyRequest
        from steamship.plugin.outputs.block_and_tag_plugin_output import BlockAndTagPluginOutput

        req = BlockifyRequest(type="file", id=self.id, plugin_instance=plugin_instance)

        return self.client.post(
            "plugin/instance/blockify",
            payload=req,
            expect=BlockAndTagPluginOutput,
            wait_on_tasks=wait_on_tasks,
        )

    def tag(
        self,
        plugin_instance: str = None,
        wait_on_tasks: List[Task] = None,
    ) -> Task[TagResponse]:
        from steamship.data.operations.tagger import TagRequest, TagResponse
        from steamship.data.plugin import PluginTargetType

        req = TagRequest(type=PluginTargetType.FILE, id=self.id, plugin_instance=plugin_instance)
        return self.client.post(
            "plugin/instance/tag", payload=req, expect=TagResponse, wait_on_tasks=wait_on_tasks
        )

    def index(self, plugin_instance: Any = None) -> EmbeddingIndex:
        """Index every block in the file.

        TODO(ted): Enable indexing the results of a tag query.
        TODO(ted): It's hard to load the EmbeddingIndexPluginInstance with just a handle because of the chain
        of things that need to be created to it to function."""

        # Preserve the prior behavior of embedding the full text of each block.
        tags = [
            Tag(text=block.text, file_id=self.id, block_id=block.id, kind="block")
            for block in self.blocks or []
        ]
        return plugin_instance.insert(tags)

    @staticmethod
    def list(client: Client) -> ListFileResponse:
        return client.post(
            "file/list",
            ListFileRequest(),
            expect=ListFileResponse,
        )


class FileQueryResponse(Response):
    files: List[File]


File.CreateRequest.update_forward_refs()
ListFileResponse.update_forward_refs()



from __future__ import annotations

from typing import Any, List, Optional

from pydantic import Field

from steamship.base.client import Client
from steamship.base.model import CamelModel
from steamship.base.request import DeleteRequest, IdentifierRequest, Request
from steamship.base.response import Response
from steamship.data.tags.tag import Tag


class BlockQueryRequest(Request):
    tag_filter_query: str


class Block(CamelModel):
    client: Client = Field(None, exclude=True)
    id: str = None
    file_id: str = None
    text: str = None
    tags: Optional[List[Tag]] = []

    class CreateRequest(Request):
        id: str = None
        file_id: str = None
        text: str = None
        tags: Optional[List[Tag.CreateRequest]] = []

    class ListRequest(Request):
        file_id: str = None

    class ListResponse(Response):
        blocks: List[Block] = []

    @staticmethod
    def get(
        client: Client,
        _id: str = None,
    ) -> Block:
        return client.post(
            "block/get",
            IdentifierRequest(id=_id),
            expect=Block,
        )

    @staticmethod
    def create(
        client: Client,
        file_id: str = None,
        text: str = None,
        tags: List[Tag.CreateRequest] = None,
    ) -> Block:
        req = Block.CreateRequest(file_id=file_id, text=text, tags=tags)
        return client.post(
            "block/create",
            req,
            expect=Block,
        )

    def delete(self) -> Block:
        return self.client.post(
            "block/delete",
            DeleteRequest(id=self.id),
            expect=Tag,
        )

    @staticmethod
    def query(
        client: Client,
        tag_filter_query: str,
    ) -> BlockQueryResponse:
        req = BlockQueryRequest(tag_filter_query=tag_filter_query)
        res = client.post(
            "block/query",
            payload=req,
            expect=BlockQueryResponse,
        )
        return res

    def index(self, embedding_plugin_instance: Any = None):
        """Index this block."""
        tags = [
            Tag(
                text=self.text,
                file_id=self.file_id,
                block_id=self.id,
                kind="block",
                start_idx=0,
                end_idx=len(self.text),
            )
        ]
        return embedding_plugin_instance.insert(tags)


class BlockQueryResponse(Response):
    blocks: List[Block]


Block.ListResponse.update_forward_refs()
Block.update_forward_refs()



from __future__ import annotations

from typing import Optional

from pydantic import Field

from steamship.base.model import CamelModel


class TrainPluginInput(CamelModel):
    """
    This is the object passed as input to a trainable operation, stored as the `input` field of a `train` task.
    """

    plugin_instance: str

    # How may epochs of trainable to perform, if relevant and supported
    training_epochs: Optional[int] = None

    # How much data to hold out for testing & reporting, if relevant and supported.
    testing_holdout_percent: Optional[float] = None

    # An optional seed for the train-test split
    test_split_seed: Optional[int] = None

    # Arbitrary key-valued data to provide to the particular `modelName` trainer.
    training_params: Optional[dict] = None

    # Arbitrary key-valued data to provide to the inference runner in the TrainPluginOutput object.
    # The trainable process will have the opportunity to amend this before writing it to the output
    inference_params: Optional[dict] = None

    # A pre-signed URL at which the trainable data can be found
    training_data_url: Optional[str] = Field(None, alias="trainingDataUrl")



from unittest.mock import patch

from assets.plugins.taggers.plugin_trainable_tagger_config import (
    TestConfig,
    TestTrainableTaggerConfigModel,
    TestTrainableTaggerConfigPlugin,
)
from steamship_tests.utils.fixtures import get_steamship_client

from steamship import File
from steamship.base import Task
from steamship.data.block import Block
from steamship.invocable.plugin_service import PluginRequest
from steamship.plugin.inputs.block_and_tag_plugin_input import BlockAndTagPluginInput
from steamship.plugin.inputs.train_plugin_input import TrainPluginInput
from steamship.plugin.request import PluginRequestContext
from steamship.plugin.trainable_model import TrainableModel

TEST_REQ = BlockAndTagPluginInput(
    file=File(
        blocks=[
            Block(
                id="ABC",
                text="Once upon a time there was a magical ship. "
                "The ship was powered by STEAM. The ship went to the moon.",
            )
        ]
    )
)
TEST_PLUGIN_REQ = PluginRequest(
    data=TEST_REQ, context=PluginRequestContext(plugin_instance_id="000")
)
TEST_PLUGIN_REQ_DICT = TEST_PLUGIN_REQ.dict()


def test_trainable_tagger():
    client = get_steamship_client()
    assert client is not None

    def load_remote(**kwargs):
        model = TestTrainableTaggerConfigModel()
        model.receive_config(TestConfig(testValue1=1, testValue2=2))
        return model

    # There isn't ACTUALLY model data to load..
    with patch.object(TrainableModel, "load_remote", load_remote):
        plugin = TestTrainableTaggerConfigPlugin(
            client=client, config={"testValue1": "foo", "testValue2": "bar"}
        )
        assert plugin.client is not None

        # Make sure plugin model gets its config while 'training'.
        plugin.train_endpoint(
            **PluginRequest(
                data=TrainPluginInput(plugin_instance="foo", training_params=None),
                status=Task(task_id="000"),
                context=PluginRequestContext(plugin_instance_id="000"),
            ).dict()
        )

        # Make sure plugin model gets its config while 'running'
        plugin.run(TEST_PLUGIN_REQ)



"""Tests the Embedding Index Snapshot functionality.

When we finish re-factoring the index code in the Engine, these tests can be deleted.
For now, it's useful to have them as-is, as they target a detail of index operation that the old interface
captures, but the new interface does not.
"""
from steamship_tests.utils.fixtures import get_steamship_client

from steamship import PluginInstance
from steamship.base import TaskState
from steamship.data.embeddings import EmbeddingIndex, IndexSnapshotRequest, IndexSnapshotResponse

_TEST_EMBEDDER = "test-embedder"


def _insert(index, items):
    for item in items:
        index.insert(item, "TestId", "TestType", [1, 2, 3])

    # Now embed
    task = index.embed()
    task.wait()
    task.refresh()
    assert task.state == TaskState.succeeded


def _snapshot(index, window_size=None):
    if window_size is None:
        task = index.create_snapshot()
        task.wait()
        task.refresh()
    else:
        # We do this manually so as not to clutter the end-user visible
        # API with debug/testing parameters
        req = IndexSnapshotRequest(index_id=index.id, window_size=window_size)
        task = index.client.post(
            "embedding-index/snapshot/create",
            req,
            expect=IndexSnapshotResponse,
        )
        task.wait()
        task.refresh()


def test_snapshot_create():
    steamship = get_steamship_client()

    plugin_instance = PluginInstance.create(steamship, plugin_handle=_TEST_EMBEDDER)
    index = EmbeddingIndex.create(
        client=steamship,
        embedder_plugin_instance_handle=plugin_instance.handle,
    )

    _insert(index, ["Oranges are orange."])
    search_results = index.search("What color are oranges?", include_metadata=True)
    search_results.wait()
    items = search_results.output.items
    assert len(items) == 1
    assert items[0].value.index_source == "index"
    assert items[0].value.value == "Oranges are orange."
    assert items[0].value.external_id == "TestId"
    assert items[0].value.external_type == "TestType"
    assert len(items[0].value.metadata) == 3

    _snapshot(index)
    search_results = index.search("What color are oranges?", include_metadata=True)
    search_results.wait()
    items = search_results.output.items
    assert len(items) == 1
    assert items[0].value.index_source == "snapshot"
    assert items[0].value.value == "Oranges are orange."
    assert items[0].value.external_id == "TestId"
    assert items[0].value.external_type == "TestType"
    assert len(items[0].value.metadata) == 3

    _insert(index, ["Apples are red."])
    search_results = index.search("What color are apples?", include_metadata=True)
    search_results.wait()
    items = search_results.output.items
    assert len(items) == 1
    assert items[0].value.index_source == "index"
    assert items[0].value.value == "Apples are red."
    assert items[0].value.external_id == "TestId"
    assert items[0].value.external_type == "TestType"
    assert len(items[0].value.metadata) == 3

    _snapshot(index)
    search_results = index.search("What color are apples?", include_metadata=True)
    search_results.wait()
    items = search_results.output.items
    assert len(items) == 1
    assert items[0].value.index_source == "snapshot"
    assert items[0].value.value == "Apples are red."
    assert items[0].value.external_id == "TestId"
    assert items[0].value.external_type == "TestType"
    assert len(items[0].value.metadata) == 3

    index.delete()
    steamship = get_steamship_client()

    index = EmbeddingIndex.create(
        client=steamship,
        embedder_plugin_instance_handle=plugin_instance.handle,
    )

    sentences = []
    for i in range(15):
        sentences.append(f"Orange number {i} is as good as the last")

    sent = "Is orange number 13 Any good?"
    _insert(index, sentences)

    search_results = index.search(sent, include_metadata=True)
    search_results.wait()
    items = search_results.output.items
    assert len(items) == 1
    assert items[0].value.index_source == "index"
    assert items[0].value.value == "Orange number 13 is as good as the last"
    assert items[0].value.external_id == "TestId"
    assert items[0].value.external_type == "TestType"
    assert len(items[0].value.metadata) == 3

    _snapshot(index, window_size=2)
    search_results = index.search(sent, include_metadata=True)
    search_results.wait()
    items = search_results.output.items
    assert len(items) == 1
    assert items[0].value.index_source == "snapshot"
    assert items[0].value.value == "Orange number 13 is as good as the last"
    assert items[0].value.external_id == "TestId"
    assert items[0].value.external_type == "TestType"
    assert len(items[0].value.metadata) == 3

    index.delete()



import json

import pytest

from steamship.client import Steamship
from steamship.data.block import Block
from steamship.data.file import File
from steamship.data.tags.tag import Tag


@pytest.mark.usefixtures("client")
def test_dict_method_should_not_include_client(client: Steamship):
    block = Block(client=client, text="Hi")
    d = block.dict()
    assert isinstance(d, dict)
    assert "client" not in d

    d = block.dict(exclude={"foo"})
    assert isinstance(d, dict)
    assert "client" not in d


@pytest.mark.usefixtures("client")
def test_serialize_client_to_json_works(client: Steamship):
    assert "use" not in client.dict().keys()
    assert "use_plugin" not in client.dict().keys()

    j = json.dumps(client.dict())  # this will fail if `use` or `use_plugin` are output by dict()
    assert j is not None


@pytest.mark.usefixtures("client")
def test_serialize_many_models(client: Steamship):
    json.dumps(Block(client=client).dict())
    json.dumps(Block(client=client, tags=[Tag(client=client)]).dict())
    json.dumps(Tag(client=client).dict())
    json.dumps(File(client=client).dict())
    json.dumps(File(client=client, tags=[Tag(client=client)]).dict())
    json.dumps(File(client=client, tags=[Tag(client=client)], blocks=[Block(client=client)]).dict())
    json.dumps(
        File(
            client=client,
            tags=[Tag(client=client)],
            blocks=[Block(client=client, tags=[Tag(client=client)])],
        ).dict()
    )



from abc import ABC, abstractmethod

from steamship.invocable import InvocableResponse, post
from steamship.invocable.plugin_service import PluginRequest, PluginService
from steamship.plugin.inputs.block_and_tag_plugin_input import BlockAndTagPluginInput
from steamship.plugin.outputs.embedded_items_plugin_output import EmbeddedItemsPluginOutput


# Note!
# =====
#
# This is the PLUGIN IMPLEMENTOR's View of an Embedder.
#
# If you are using the Steamship Client, you probably want steamship.client.operations.embedder instead
# of this file.
#
class Embedder(PluginService[BlockAndTagPluginInput, EmbeddedItemsPluginOutput], ABC):
    @abstractmethod
    def run(
        self, request: PluginRequest[BlockAndTagPluginInput]
    ) -> InvocableResponse[EmbeddedItemsPluginOutput]:
        raise NotImplementedError()

    @post("tag")
    def run_endpoint(self, **kwargs) -> InvocableResponse[EmbeddedItemsPluginOutput]:
        """Exposes the Embedder's `run` operation to the Steamship Engine via the expected HTTP path POST /tag"""
        return self.run(PluginRequest[BlockAndTagPluginInput](**kwargs))



from __future__ import annotations

import time
from typing import Any, Dict, Generic, List, Optional, Set, Type, TypeVar

from pydantic import BaseModel, Field

from steamship.base.error import SteamshipError
from steamship.base.model import CamelModel, GenericCamelModel
from steamship.base.request import DeleteRequest, IdentifierRequest, Request
from steamship.utils.metadata import metadata_to_str, str_to_metadata

T = TypeVar("T")


class CreateTaskCommentRequest(Request):
    task_id: str
    external_id: str = None
    external_type: str = None
    external_group: str = None
    metadata: str = None


class ListTaskCommentRequest(Request):
    task_id: str = None
    external_id: str = None
    external_type: str = None
    external_group: str = None


class TaskComment(CamelModel):
    client: Client = Field(None, exclude=True)
    id: str = None
    user_id: str = None
    task_id: str = None
    external_id: str = None
    external_type: str = None
    external_group: str = None
    metadata: Any = None
    created_at: str = None

    def __init__(self, **kwargs):
        kwargs["metadata"] = str_to_metadata(kwargs.get("metadata"))
        super().__init__(**kwargs)

    @classmethod
    def parse_obj(cls: Type[BaseModel], obj: Any) -> BaseModel:
        # TODO (enias): This needs to be solved at the engine side
        obj = obj["taskComment"] if "taskComment" in obj else obj
        return super().parse_obj(obj)

    @staticmethod
    def create(
        client: Client,
        task_id: str = None,
        external_id: str = None,
        external_type: str = None,
        external_group: str = None,
        metadata: Any = None,
    ) -> TaskComment:
        req = CreateTaskCommentRequest(
            taskId=task_id,
            external_id=external_id,
            external_type=external_type,
            externalGroup=external_group,
            metadata=metadata_to_str(metadata),
        )
        return client.post(
            "task/comment/create",
            req,
            expect=TaskComment,
        )

    @staticmethod
    def list(
        client: Client,
        task_id: str = None,
        external_id: str = None,
        external_type: str = None,
        external_group: str = None,
    ) -> TaskCommentList:
        req = ListTaskCommentRequest(
            taskId=task_id,
            external_id=external_id,
            external_type=external_type,
            externalGroup=external_group,
        )
        return client.post(
            "task/comment/list",
            req,
            expect=TaskCommentList,
        )

    def delete(self) -> TaskComment:
        req = DeleteRequest(id=self.id)
        return self.client.post(
            "task/comment/delete",
            req,
            expect=TaskComment,
        )


class TaskCommentList(CamelModel):
    comments: List[TaskComment]


class TaskState:
    waiting = "waiting"
    running = "running"
    succeeded = "succeeded"
    failed = "failed"


class TaskType:
    internal_api = "internalApi"
    train = "train"
    infer = "infer"


class TaskRunRequest(Request):
    task_id: str


class TaskStatusRequest(Request):
    task_id: str


class Task(GenericCamelModel, Generic[T]):
    """Encapsulates a unit of asynchronously performed work."""

    client: Client = Field(None, exclude=True)  # Steamship client

    task_id: str = None  # The id of this task
    user_id: str = None  # The user who requested this task
    workspace_id: str = None  # The workspace in which this task is executing
    expect: Type = None  # Type of the expected output once the output is complete

    input: str = None  # The input provided to the task
    output: T = None  # The output of the task
    state: str = None  # A value in class TaskState

    status_message: str = None  # User-facing message concerning task status
    status_suggestion: str = None  # User-facing suggestion concerning error remediation
    status_code: str = None  # User-facing error code for support assistance
    status_created_on: str = None  # When the status fields were last set

    task_type: str = None  # A value in class TaskType; for internal routing
    task_executor: str = None  #
    task_created_on: str = None  # When the task object was created
    task_last_modified_on: str = None  # When the task object was last modified

    # Long Running Plugin Support
    # The `remote_status_*` fields govern how Steamship Plugins can communicate long-running work back to the engine.
    # If instead of sending data, the plugin sends a status with these fields set, the engine will begin polling for
    # updates, echoing the contents of these fields back to the plugin to communicate, e.g., the jobId of the work
    # being checked. When the work is complete, simply respond with the Response `data` field set as per usual.
    remote_status_input: Optional[
        Dict
    ] = None  # For re-hydrating state in order to check remote status.
    remote_status_output: Optional[
        Dict
    ] = None  # For reporting structured JSON state for error diagnostics.
    remote_status_message: str = None  # User facing message string to report on remote status.

    assigned_worker: str = None  # The worker assigned to complete this task
    started_at: str = None  # When the work on this task began

    max_retries: int = None  # The maximum number of retries allowed for this task
    retries: int = None  # The number of retries already used.

    def as_error(self) -> SteamshipError:
        return SteamshipError(
            message=self.status_message, suggestion=self.status_suggestion, code=self.status_code
        )

    @classmethod
    def parse_obj(cls: Type[BaseModel], obj: Any) -> Task:
        obj = obj["task"] if "task" in obj else obj
        return super().parse_obj(obj)

    @staticmethod
    def get(
        client,
        _id: str = None,
        handle: str = None,
    ) -> Task:
        return client.post(
            "task/get",
            IdentifierRequest(id=_id, handle=handle),
            expect=Task,
        )

    def update(self, other: Optional[Task] = None):
        """Incorporates a `Task` into this object."""
        other = other or Task()
        for k, v in other.__dict__.items():
            self.__dict__[k] = v

    def add_comment(
        self,
        external_id: str = None,
        external_type: str = None,
        external_group: str = None,
        metadata: Any = None,
    ) -> TaskComment:
        return TaskComment.create(
            client=self.client,
            task_id=self.task_id,
            external_id=external_id,
            external_type=external_type,
            external_group=external_group,
            metadata=metadata,
        )

    def post_update(self, fields: Set[str] = None) -> Task:
        """Updates this task in the Steamship Engine."""
        if not isinstance(fields, set):
            raise RuntimeError(f'Unexpected type of "fields": {type(fields)}. Expected type set.')
        body = self.dict(by_alias=True, include={*fields, "task_id"})
        return self.client.post("task/update", body, expect=Task)

    def wait(self, max_timeout_s: float = 60, retry_delay_s: float = 1):
        """Polls and blocks until the task has succeeded or failed (or timeout reached)."""
        t0 = time.perf_counter()
        while time.perf_counter() - t0 < max_timeout_s and self.state not in (
            TaskState.succeeded,
            TaskState.failed,
        ):
            self.refresh()
            time.sleep(retry_delay_s)

        # If the task did not complete within the timeout, throw an error
        if self.state not in (TaskState.succeeded, TaskState.failed):
            raise SteamshipError(
                message=f"Task {self.task_id} did not complete within requested timeout of {max_timeout_s}s"
            )

    def refresh(self):
        req = TaskStatusRequest(taskId=self.task_id)
        # TODO (enias): A status call can return both data and task
        # In this case both task and data will include the output (one is string serialized, the other is parsed)
        # Ideally task status only returns the status, not the full output object
        resp = self.client.post("task/status", payload=req, expect=self.expect)
        self.update(resp)


from .client import Client  # noqa: E402

Task.update_forward_refs()
TaskComment.update_forward_refs()



from __future__ import annotations

import io
from typing import Any, Optional, Type, Union

from pydantic import BaseModel

from steamship.base import MimeTypes
from steamship.base.model import CamelModel
from steamship.utils.binary_utils import flexi_create


class RawDataPluginOutput(CamelModel):
    """Represents mime-typed raw data (or a URL pointing to raw data) that can be returned to the engine.

    As a few examples, you can return:
    - Raw text: RawDataPluginOutput(string=raw_text, MimeTypes.TXT)
    - Markdown text: RawDataPluginOutput(string=markdown_text, MimeTypes.MKD)
    - A PNG image: RawDataPluginOutput(bytes=png_bytes, MimeTypes.PNG)
    - A JSON-serializable Dataclass: RawDataPluginOutput(json=dataclass, MimeTypes.JSON)
    - Steamship Blocks: RawDataPluginOutput(json=file, MimeTypes.STEAMSHIP_BLOCK_JSON)
    - Data uploaded to a pre-signed URL: RawDataPluginOutput(url=presigned_url, MimeTypes.TXT)

    The `data` field of this object will ALWAYS be Base64 encoded by the constructor. This ensures that the object
    is always trivially JSON-serializable over the wire, no matter what it contains.

    The `mimeType` field of this object should always be filled in if known. The Steamship Engine makes use of it
    to proactively select defaults for handling the data returned.
    """

    data: Optional[str] = None  # Note: This is **always** Base64 encoded.
    mime_type: Optional[str] = None

    def __init__(
        self,
        base64string: str = None,
        string: str = None,
        _bytes: Union[bytes, io.BytesIO] = None,
        json: Any = None,
        mime_type: str = None,
        **kwargs,
    ):
        super().__init__()

        if base64string is not None:
            self.data = base64string
            self.mime_type = mime_type or MimeTypes.BINARY
        else:
            # Base64-encode the data field.
            self.data, self.mime_type, encoding = flexi_create(
                base64string=base64string,
                string=string,
                json=json,
                _bytes=_bytes,
                mime_type=mime_type,
                force_base64=True,
            )

    @classmethod
    def parse_obj(cls: Type[BaseModel], obj: Any) -> BaseModel:
        obj["base64string"] = obj.get("data")
        return super().parse_obj(obj)



import logging
import tempfile
from pathlib import Path
from typing import ClassVar, Optional

from steamship import Steamship, SteamshipError
from steamship.base.client import Client
from steamship.base.model import CamelModel
from steamship.data.workspace import SignedUrl, Workspace
from steamship.utils.signed_urls import download_from_signed_url, upload_to_signed_url
from steamship.utils.zip_archives import unzip_folder, zip_folder


class ModelCheckpoint(CamelModel):
    # The default model checkpoint handle unless one is provided.
    DEFAULT_HANDLE: ClassVar[str] = "default"

    """Represents the saved state of a trained PluginInstance.
    """
    client: Client
    workspace: Optional[Workspace] = None
    plugin_instance_id: str

    parent_directory: Optional[Path] = None  # e.g. /tmp
    handle: str = None  # The handle of this ModelCheckpoint.
    plugin_instance_id: str = None  #

    def __init__(
        self,
        client: Steamship,
        parent_directory: Optional[Path] = None,
        handle: str = DEFAULT_HANDLE,
        plugin_instance_id: str = None,
    ):
        super().__init__(
            client=client,
            parent_directory=parent_directory,
            plugin_instance_id=plugin_instance_id,
            handle=handle or ModelCheckpoint.DEFAULT_HANDLE,
        )

        if self.plugin_instance_id is None:
            raise SteamshipError("Null plugin_instance_id provided ModelCheckpoint")

        self.workspace = client.get_workspace()

        if parent_directory is None:
            # TODO(ted): We may want to not use a tempdir so that we can cache it.
            self.parent_directory = Path(tempfile.mkdtemp())

        # Create the folder path on disk.
        logging.info(f"Making sure Checkpoint path exists: {self.folder_path_on_disk()}")
        self.folder_path_on_disk().mkdir(parents=True, exist_ok=True)

    def folder_path_on_disk(self) -> Path:
        """Returns the path to this checkpoint on the local disk.

        On disk, the model checkpoint is the folder:
            `{parent_directory}/{checkpoint_handle}/`
        """
        return self.parent_directory / Path(self.handle)

    def archive_path_on_disk(self) -> Path:
        """Returns the path to the checkpoint archive on disk.

        On disk, the model checkpoint is the folder:
            `{parent_directory}/{checkpoint_handle}.zip`
        """
        return self.parent_directory / Path(f"{self.handle}.zip")

    def archive_path_in_steamship(self, as_handle: str = None) -> str:
        """Returns the path to the checkpoint archive on Steamship.

        On steamship, the checkpoint is archived in the Workspace's PluginInstance bucket as:
        `{plugin_instance_bucket}/{plugin_instance_id}/{checkpoint_handle}.zip`

        Here we only return the following path since the bucket is specified separately
        in the required Steamship API calls: `{plugin_instance_id}/{checkpoint_handle}.zip`
        """
        return f"{self.plugin_instance_id}/{as_handle or self.handle}.zip"

    def download_model_bundle(self) -> Path:
        """Download's the model from Steamship and unzips to `parent_directory`"""
        download_resp = self.workspace.create_signed_url(
            SignedUrl.Request(
                bucket=SignedUrl.Bucket.PLUGIN_DATA,
                filepath=self.archive_path_in_steamship(),
                operation=SignedUrl.Operation.READ,
            )
        )
        if not download_resp or not download_resp.signed_url:
            raise SteamshipError(
                message=f"Received empty Signed URL for model download of '{self.handle}."
            )
        download_from_signed_url(download_resp.signed_url, to_file=self.archive_path_on_disk())
        unzip_folder(self.archive_path_on_disk(), into_folder=self.folder_path_on_disk())
        if not download_resp or not download_resp.signed_url:
            raise SteamshipError(
                message=f"Received empty Signed URL for model download of '{self.handle}."
            )
        download_from_signed_url(download_resp.signed_url, to_file=self.archive_path_on_disk())
        unzip_folder(self.archive_path_on_disk(), into_folder=self.folder_path_on_disk())
        return self.folder_path_on_disk()

    def _upload_model_zip(self, as_handle: str = None):
        """Assumes a pre-zipped model, uploads to the requested zip.

        This is an internal function. Please use upload_model_bundle as an caller."""
        logging.info(f"ModelCheckpoint:_upload_model_zip - handle={as_handle}")
        signed_url_resp = self.workspace.create_signed_url(
            SignedUrl.Request(
                bucket=SignedUrl.Bucket.PLUGIN_DATA,
                filepath=self.archive_path_in_steamship(as_handle=as_handle),
                operation=SignedUrl.Operation.WRITE,
            )
        )

        if not signed_url_resp:
            raise SteamshipError(
                message="Empty result on Signed URL request while uploading model checkpoint"
            )
        if not signed_url_resp.signed_url:
            raise SteamshipError(
                message="Empty signedUrl on Signed URL request while uploading model checkpoint"
            )

        upload_to_signed_url(signed_url_resp.signed_url, filepath=self.archive_path_on_disk())

    def upload_model_bundle(self, set_as_default: bool = True):
        """Zips and uploads the Model to steamship"""
        logging.info("ModelCheckpoint:upload_model_bundle")
        zip_folder(self.folder_path_on_disk(), into_file=self.archive_path_on_disk())
        self._upload_model_zip()

        if set_as_default:
            # For simplicity, we'll assume the checkpoint named `default` is the one to be loaded unless otherwise
            # specified. This means that we need to double-upload some checkpoints:
            # - Once under the actual checkpoint name (e.g. `epoch-10`)
            # - Again under the name: default
            self._upload_model_zip(as_handle=ModelCheckpoint.DEFAULT_HANDLE)
